-- Copyright (C) 2013 Nicholas Carlson
--
-- Permission is hereby granted, free of charge, to any person obtaining a
-- copy of this software and associated documentation files (the "Software"),
-- to deal in the Software without restriction, including without limitation
-- the rights to use, copy, modify, merge, publish, distribute, sublicense,
-- and/or sell copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
-- DEALINGS IN THE SOFTWARE.

local colorspace = require 'lib.colorspace'
local vec2 = require 'lib.math.vector'
local vec3 = require 'lib.math.vec3'
local vec4 = require 'lib.math.vec4'
local mat4 = require 'lib.math.mat4'

local model = mat4.identity()
local view = mat4.lookat( vec3( 0, 1, 2 ),
                          vec3( 0, 0, 0 ),
                          vec3( 0, 1, 0 ) )
local projection = mat4.perspectiveFov( 55, gWindow.width/gWindow.height, 0.1, 100 )

local red =     vec2(colorspace.rgbToXY(255, 000, 000))
local green =   vec2(colorspace.rgbToXY(000, 255, 000))
local blue =    vec2(colorspace.rgbToXY(000, 000, 255))
local yellow =  vec2(colorspace.rgbToXY(255, 255, 000))
local cyan =    vec2(colorspace.rgbToXY(000, 255, 255))
local pink =    vec2(colorspace.rgbToXY(255, 000, 255))



function buildCube(imgPath)
  local left = {
    { yellow.x, yellow.y,   0, 0,   000, 255, 000, 255 },
    { yellow.x, yellow.y,   0, 1,   000, 000, 000, 255 },
    { yellow.x, yellow.y,   1, 1,   000, 000, 255, 255 },
    { yellow.x, yellow.y,   1, 0,   000, 255, 255, 255 },
  }

  local right = {
    { blue.x, blue.y,   0, 0,   255, 255, 255, 255 },
    { blue.x, blue.y,   0, 1,   255, 000, 255, 255 },
    { blue.x, blue.y,   1, 1,   255, 000, 000, 255 },
    { blue.x, blue.y,   1, 0,   255, 255, 000, 255 },
  }

  local top = {
    { pink.x, pink.y,   0, 0,   000, 255, 000, 255 },
    { pink.x, pink.y,   0, 1,   000, 255, 255, 255 },
    { pink.x, pink.y,   1, 1,   255, 255, 255, 255 },
    { pink.x, pink.y,   1, 0,   255, 255, 000, 255 },
  }

  local bottom = {
    { cyan.x, cyan.y,   0, 0,   000, 000, 255, 255 },
    { cyan.x, cyan.y,   0, 1,   000, 000, 000, 255 },
    { cyan.x, cyan.y,   1, 1,   255, 000, 000, 255 },
    { cyan.x, cyan.y,   1, 0,   255, 000, 255, 255 },
  }

  local front = {
    { green.x, green.y,   0, 0,   000, 255, 255, 255 },
    { green.x, green.y,   0, 1,   000, 000, 255, 255 },
    { green.x, green.y,   1, 1,   255, 000, 255, 255 },
    { green.x, green.y,   1, 0,   255, 255, 255, 255 },
  }

  local back = {
    { red.x, red.y,   0, 0,   255, 255, 000, 255 },
    { red.x, red.y,   0, 1,   255, 000, 000, 255 },
    { red.x, red.y,   1, 1,   000, 000, 000, 255 },
    { red.x, red.y,   1, 0,   000, 255, 000, 255 },
  }


  local tex = love.graphics.newImage(imgPath);

  local cube = {
    draw = function (self)
      love.graphics.draw(self.left, 0, 0)
      love.graphics.draw(self.right, 0, 0)
      love.graphics.draw(self.top, 0, 0)
      love.graphics.draw(self.bottom, 0, 0)
      love.graphics.draw(self.front, 0, 0)
      love.graphics.draw(self.back, 0, 0)
    end;

    left = love.graphics.newMesh(left, tex);
    right = love.graphics.newMesh(right, tex);
    top = love.graphics.newMesh(top, tex);
    bottom = love.graphics.newMesh(bottom, tex);
    front = love.graphics.newMesh(front, tex);
    back = love.graphics.newMesh(back, tex);
  }

  return cube
end

function love.load()
  love.graphics.setDefaultFilter("nearest", "nearest")

  vs = love.filesystem.read("assets/cube.vert")
  fs = love.filesystem.read("assets/cube.frag")

  shader = love.graphics.newShader(vs, fs)
  love.graphics.setShader(shader)

  shader:send("Projection", projection:lovemat())
  shader:send("View", view:lovemat())
  shader:send("Model", model:lovemat())

  cube = buildCube("assets/stone.png")
end

function love.update(dt)
  local rock = mat4.rotate(math.sin(love.timer.getTime()), 1, 0, 0) * mat4.rotate(math.cos(love.timer.getTime()), 1, 0, 0)
  local rotate = mat4.rotate(love.timer.getTime()*2, 0, 1, 0)
  local translate = mat4.translate(-0.5, -0.5, -0.5)
  local rotated = model * rotate * rock * translate
  shader:send("Model", rotated:lovemat())
end

function love.quit()
end

function love.draw()
  cube:draw()
end

function love.visible(v)
end

function love.resize(w,h)
  local p = mat4.perspectiveFov( 55, w/h, 0.1, 100 )  
  shader:send("Projection", projection:lovemat())
  shader:send("View", view:lovemat())
  shader:send("Model", model:lovemat())
end

function love.visible(v)
end

function love.keypressed(b,u)
  if b == "escape" then
      love.event.push("quit")
  end
end

function love.keyreleased(b)
end

function love.mousepressed(x,y,b)
end

function love.mousereleased(x,y,b)
end

function love.joystickpressed(j,b)
end

function love.joystickreleased(j,b)
end

function love.joystickaxis(j,a,v)
end

function love.joystickhat(j,h,v)
end